# Copyright (c) 2021 Philip May, Deutsche Telekom AG
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT


"""Transformers tools."""
import sklearn
import torch


class LabeledDataset(torch.utils.data.Dataset):
    """Dataset with labes."""

    def __init__(self, encodings, labels):
        self.encodings = encodings
        self.labels = labels

    def __getitem__(self, idx):
        item = {key: torch.tensor(val[idx]) for key, val in self.encodings.items()}
        item["labels"] = torch.tensor(self.labels[idx])
        return item

    def __len__(self):
        return len(self.labels)


class KFoldLabeledDataset:
    """Utility to do k-fold cross-validation on ``LabeledDataset``."""

    def __init__(self, n_splits=7, n_repeats=1, random_state=None):
        self.n_splits = n_splits
        self.n_repeats = n_repeats
        self.random_state = random_state

    def split(self, labeled_dataset, stratification_labels=None):
        """Generates data splits of training and test set."""
        idxs = list(range(len(labeled_dataset)))
        if stratification_labels is None:  # no stratification wanted
            k_fold = sklearn.model_selection.RepeatedKFold(
                n_splits=self.n_splits,
                n_repeats=self.n_repeats,
                random_state=self.random_state,
            )
            k_fold_split = k_fold.split(idxs)
        else:  # stratification wanted
            k_fold = sklearn.model_selection.RepeatedStratifiedKFold(
                n_splits=self.n_splits,
                n_repeats=self.n_repeats,
                random_state=self.random_state,
            )
            k_fold_split = k_fold.split(idxs, stratification_labels)

        for train_idxs, test_idxs in k_fold_split:
            train = torch.utils.data.Subset(labeled_dataset, train_idxs)
            test = torch.utils.data.Subset(labeled_dataset, test_idxs)
            yield train, test



# Copyright (c) 2021 Philip May, Deutsche Telekom AG
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

"""Transformers callbacks.

Also see :class:`transformers.TrainerCallback`.
"""

import logging
import shutil
import warnings
from pathlib import Path
from typing import Any, Dict, Optional

from ml_cloud_tools.s3 import copy_dir_to_s3_dir, copy_s3_dir_to_dir, list_s3_files
from transformers import TrainerCallback, TrainerControl, TrainerState, TrainingArguments
from transformers.trainer_utils import get_last_checkpoint


_logger = logging.getLogger(__name__)


class S3CheckpointSyncCallback(TrainerCallback):
    """Save and recover checkpoints by using S3."""

    def __init__(
        self,
        s3_dir_name: str,
        s3_bucket_name: Optional[str] = None,
        s3_kwargs: Optional[Dict[str, Any]] = None,
    ):
        self.s3_dir_name = s3_dir_name
        self.s3_bucket_name = s3_bucket_name
        self.s3_kwargs = s3_kwargs

    def on_init_end(
        self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs
    ):
        """Event called at the end of the initialization of the :class:`~transformers.Trainer`."""
        # TODO: only copy last CP
        source_s3_dir_path = Path(self.s3_dir_name) / Path(args.output_dir).name
        files_on_s3 = list_s3_files(
            source_s3_dir_path.as_posix(), s3_bucket_name=self.s3_bucket_name
        )
        if len(files_on_s3) > 0:
            target_dir_path = Path(args.output_dir).resolve().parent
            copy_s3_dir_to_dir(
                source_s3_dir_path.as_posix(),
                target_dir_path.as_posix(),
                s3_bucket_name=self.s3_bucket_name,
                s3_kwargs=self.s3_kwargs,
            )

    def on_save(
        self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs
    ):
        """Event called after a checkpoint save."""
        last_checkpoint_dir = get_last_checkpoint(args.output_dir)
        if last_checkpoint_dir is not None:
            source_dir_path = Path(last_checkpoint_dir)
            target_s3_dir_path = Path(self.s3_dir_name) / Path(args.output_dir).name
            copy_dir_to_s3_dir(
                source_dir_path.as_posix(),
                target_s3_dir_path.as_posix(),
                s3_bucket_name=self.s3_bucket_name,
                s3_kwargs=self.s3_kwargs,
            )

        # remove local checkpoint directory
        try:
            shutil.rmtree(args.output_dir)
        except Exception as e:
            error_msg = "Exception raised while deleting checkpoints! Exception: {}".format(e)
            _logger.error(error_msg, exc_info=True)
            warnings.warn(error_msg, RuntimeWarning)



# Copyright (c) 2021 Philip May
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

"""Transformer-Tools main package."""

from transformer_tools.transformers import LabeledDataset
from transformer_tools.version import __version__


__all__ = ["LabeledDataset", "__version__"]
